@@ -14,6 +14,7 @@
  *	it does cope with multiple interpreters in multiple threads.
  *	
  * Copyright (c) 2001-2004 Vince Darley.
+ * Copyright (c) 2006 ActiveState Software Inc.
  * 
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -38,6 +39,18 @@
 #define TCL_GLOB_TYPE_MOUNT		(1<<7)
 #endif
 
+/*
+ * tclvfs will return this code instead of TCL_OK/ERROR/etc. to propagate
+ * through the Tcl_Eval* calls to indicate a posix error has been raised by
+ * some vfs implementation.  -1 is what Tcl expects, adopts from posix's
+ * standard error value.
+ */
+#define TCLVFS_POSIXERROR (-1)
+
+#ifndef CONST86
+#define CONST86
+#endif
+
 /*
  * Only the _Init function is exported.
  */
@@ -51,37 +64,6 @@ EXTERN int Vfs_Init _ANSI_ARGS_((Tcl_Interp*));
 static void Vfs_AddVolume    _ANSI_ARGS_((Tcl_Obj*));
 static int  Vfs_RemoveVolume _ANSI_ARGS_((Tcl_Obj*));
 
-/* 
- * Stores the list of volumes registered with the vfs (and therefore
- * also registered with Tcl).  It is maintained as a valid Tcl list at
- * all times, or NULL if there are none (we don't keep it as an empty
- * list just as a slight optimisation to improve Tcl's efficiency in
- * determining whether paths are absolute or relative).
- * 
- * We keep a refCount on this object whenever it is non-NULL.
- */
-static Tcl_Obj *vfsVolumes = NULL;
-
-/* 
- * Declare a mutex for thread-safety of modification of the
- * list of vfs volumes.
- */
-TCL_DECLARE_MUTEX(vfsVolumesMutex)
-
-/* 
- * Stores a script to evaluate when an internal error is detected in
- * a tclvfs implementation.  This is most useful for debugging.
- * 
- * When it is not NULL we keep a refCount on it.
- */
-static Tcl_Obj *internalErrorScript = NULL;
-
-/* 
- * Declare a mutex for thread-safety of modification of the
- * internal error script.
- */
-TCL_DECLARE_MUTEX(internalErrorMutex)
-
 /*
  * struct Vfs_InterpCmd --
  * 
@@ -262,12 +244,31 @@ typedef struct VfsMount {
     struct VfsMount* nextMount;
 } VfsMount;
 
-static VfsMount* listOfMounts = NULL;
-/* 
- * Declare a mutex for thread-safety of modification of the
- * list of vfs mounts.
+#define TCL_TSD_INIT(keyPtr)	(ThreadSpecificData *)Tcl_GetThreadData((keyPtr), sizeof(ThreadSpecificData))
+
+/*
+ * Declare a thread-specific list of vfs mounts and volumes.
+ *
+ * Stores the list of volumes registered with the vfs (and therefore
+ * also registered with Tcl).  It is maintained as a valid Tcl list at
+ * all times, or NULL if there are none (we don't keep it as an empty
+ * list just as a slight optimisation to improve Tcl's efficiency in
+ * determining whether paths are absolute or relative).
+ *
+ * We keep a refCount on this object whenever it is non-NULL.
+ *
+ * internalErrorScript is evaluated when an internal error is detected in
+ * a tclvfs implementation.  This is most useful for debugging.
+ *
+ * When it is not NULL we keep a refCount on it.
  */
-TCL_DECLARE_MUTEX(vfsMountsMutex)
+
+typedef struct ThreadSpecificData {
+    VfsMount *listOfMounts;
+    Tcl_Obj *vfsVolumes;
+    Tcl_Obj *internalErrorScript;
+} ThreadSpecificData;
+static Tcl_ThreadDataKey dataKey;
 
 /* We might wish to consider exporting these in the future */
 
@@ -285,6 +286,7 @@ static void            Vfs_RegisterWithInterp _ANSI_ARGS_((Tcl_Interp*));
 static VfsNativeRep*   VfsGetNativePath(Tcl_Obj* pathPtr);
 static Tcl_CloseProc   VfsCloseProc;
 static void            VfsExitProc(ClientData clientData);
+static void            VfsThreadExitProc(ClientData clientData);
 static Tcl_Obj*	       VfsFullyNormalizePath(Tcl_Interp *interp, 
 				             Tcl_Obj *pathPtr);
 static Tcl_Obj*        VfsBuildCommandForPath(Tcl_Interp **iRef, 
@@ -342,7 +344,7 @@ Vfs_Init(interp)
 
 #ifndef PACKAGE_VERSION
     /* keep in sync with actual version */
-#define PACKAGE_VERSION "1.3"
+#define PACKAGE_VERSION "1.4"
 #endif
     if (Tcl_PkgProvide(interp, "vfs", PACKAGE_VERSION) == TCL_ERROR) {
         return TCL_ERROR;
@@ -396,6 +398,7 @@ Vfs_RegisterWithInterp(interp)
     if (vfsAlreadyRegistered == NULL) {
 	Tcl_FSRegister((ClientData)1, &vfsFilesystem);
 	Tcl_CreateExitHandler(VfsExitProc, (ClientData)NULL);
+	Tcl_CreateThreadExitHandler(VfsThreadExitProc, NULL);
     }
 }
    
@@ -465,6 +468,7 @@ Vfs_AddMount(mountPoint, isVolume, interp, mountCmd)
     char *strRep;
     int len;
     VfsMount *newMount;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
     
     if (mountPoint == NULL || interp == NULL || mountCmd == NULL) {
 	return TCL_ERROR;
@@ -497,10 +501,8 @@ Vfs_AddMount(mountPoint, isVolume, interp, mountCmd)
     newMount->isVolume = isVolume;
     Tcl_IncrRefCount(mountCmd);
     
-    Tcl_MutexLock(&vfsMountsMutex);
-    newMount->nextMount = listOfMounts;
-    listOfMounts = newMount;
-    Tcl_MutexUnlock(&vfsMountsMutex);
+    newMount->nextMount = tsdPtr->listOfMounts;
+    tsdPtr->listOfMounts = newMount;
 
     if (isVolume) {
 	Vfs_AddVolume(mountPoint);
@@ -542,6 +544,7 @@ Vfs_RemoveMount(mountPoint, interp)
     /* These two are only used if mountPoint is non-NULL */
     char *strRep = NULL;
     int len = 0;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
     
     VfsMount *mountIter;
     /* Set to NULL just to avoid warnings */
@@ -550,9 +553,8 @@ Vfs_RemoveMount(mountPoint, interp)
     if (mountPoint != NULL) {
 	strRep = Tcl_GetStringFromObj(mountPoint, &len);
     }
-       
-    Tcl_MutexLock(&vfsMountsMutex);
-    mountIter = listOfMounts;
+
+    mountIter = tsdPtr->listOfMounts;
     
     while (mountIter != NULL) {
 	if ((interp == mountIter->interpCmd.interp) 
@@ -560,8 +562,8 @@ Vfs_RemoveMount(mountPoint, interp)
 		(mountIter->mountLen == len && 
 		 !strcmp(mountIter->mountPoint, strRep)))) {
 	    /* We've found the mount. */
-	    if (mountIter == listOfMounts) {
-		listOfMounts = mountIter->nextMount;
+	    if (mountIter == tsdPtr->listOfMounts) {
+		tsdPtr->listOfMounts = mountIter->nextMount;
 	    } else {
 		lastMount->nextMount = mountIter->nextMount;
 	    }
@@ -581,13 +583,11 @@ Vfs_RemoveMount(mountPoint, interp)
 	    Tcl_DecrRefCount(mountIter->interpCmd.mountCmd);
 	    ckfree((char*)mountIter);
 	    Tcl_FSMountsChanged(&vfsFilesystem);
-	    Tcl_MutexUnlock(&vfsMountsMutex);
 	    return TCL_OK;
 	}
 	lastMount = mountIter;
 	mountIter = mountIter->nextMount;
     }
-    Tcl_MutexUnlock(&vfsMountsMutex);
     return TCL_ERROR;
 }
 
@@ -620,6 +620,7 @@ Vfs_FindMount(pathMount, mountLen)
 {
     VfsMount *mountIter;
     char *mountStr;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
     
     if (pathMount == NULL) {
 	return NULL;
@@ -631,19 +632,15 @@ Vfs_FindMount(pathMount, mountLen)
 	mountStr = Tcl_GetString(pathMount);
     }
 
-    Tcl_MutexLock(&vfsMountsMutex);
-
-    mountIter = listOfMounts;
+    mountIter = tsdPtr->listOfMounts;
     while (mountIter != NULL) {
 	if (mountIter->mountLen == mountLen && 
 	  !strncmp(mountIter->mountPoint, mountStr, (size_t)mountLen)) {
 	    Vfs_InterpCmd *ret = &mountIter->interpCmd;
-	    Tcl_MutexUnlock(&vfsMountsMutex);
 	    return ret;
 	}
 	mountIter = mountIter->nextMount;
     }
-    Tcl_MutexUnlock(&vfsMountsMutex);
     return NULL;
 }
 
@@ -663,18 +660,16 @@ Vfs_ListMounts(void)
 {
     VfsMount *mountIter;
     Tcl_Obj *res = Tcl_NewObj();
-
-    Tcl_MutexLock(&vfsMountsMutex);
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     /* Build list of mounts */
-    mountIter = listOfMounts;
+    mountIter = tsdPtr->listOfMounts;
     while (mountIter != NULL) {
 	Tcl_Obj* mount = Tcl_NewStringObj(mountIter->mountPoint, 
 					  mountIter->mountLen);
 	Tcl_ListObjAppendElement(NULL, res, mount);
 	mountIter = mountIter->nextMount;
     }
-    Tcl_MutexUnlock(&vfsMountsMutex);
     return res;
 }
 
@@ -704,6 +699,7 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
     Tcl_Obj	*CONST objv[];
 {
     int index;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
     static CONST char *optionStrings[] = {
 	"info", "internalerror", "mount", "unmount", 
@@ -733,28 +729,24 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 	    }
 	    if (objc == 2) {
 	        /* Return the current script */
-		Tcl_MutexLock(&internalErrorMutex);
-		if (internalErrorScript != NULL) {
-		    Tcl_SetObjResult(interp, internalErrorScript);
+		if (tsdPtr->internalErrorScript != NULL) {
+		    Tcl_SetObjResult(interp, tsdPtr->internalErrorScript);
 		}
-		Tcl_MutexUnlock(&internalErrorMutex);
 	    } else {
 		/* Set the script */
 		int len;
-		Tcl_MutexLock(&internalErrorMutex);
-		if (internalErrorScript != NULL) {
-		    Tcl_DecrRefCount(internalErrorScript);
+		if (tsdPtr->internalErrorScript != NULL) {
+		    Tcl_DecrRefCount(tsdPtr->internalErrorScript);
 		}
 		Tcl_GetStringFromObj(objv[2], &len);
 		if (len == 0) {
 		    /* Clear our script */
-		    internalErrorScript = NULL;
+		    tsdPtr->internalErrorScript = NULL;
 		} else {
 		    /* Set it */
-		    internalErrorScript = objv[2];
-		    Tcl_IncrRefCount(internalErrorScript);
+		    tsdPtr->internalErrorScript = objv[2];
+		    Tcl_IncrRefCount(tsdPtr->internalErrorScript);
 		}
-		Tcl_MutexUnlock(&internalErrorMutex);
 	    }
 	    return TCL_OK;
 	}
@@ -768,7 +760,12 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 		return TCL_ERROR;
 	    }
 	    Tcl_SetErrno(posixError);
-	    return -1;
+	    /*
+	     * This special error code propagate to the Tcl_Eval* calls in
+	     * other parts of the vfs C code to indicate a posix error
+	     * being raised by some vfs implementation.
+	     */
+	    return TCLVFS_POSIXERROR;
 	}
 	case VFS_NORMALIZE: {
 	    Tcl_Obj *path;
@@ -806,7 +803,7 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 		int retVal;
 		path = VfsFullyNormalizePath(interp, objv[2]);
 		retVal = Vfs_AddMount(path, 0, interp, objv[3]);
-		if (path != NULL) Tcl_DecrRefCount(path);
+		if (path != NULL) { Tcl_DecrRefCount(path); }
 		return retVal;
 	    }
 	    break;
@@ -864,14 +861,14 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 
 /* Handle an error thrown by a tcl vfs implementation */
 static void
-VfsInternalError(Tcl_Interp* interp) {
+VfsInternalError(Tcl_Interp* interp)
+{
     if (interp != NULL) {
-	Tcl_MutexLock(&internalErrorMutex);
-	if (internalErrorScript != NULL) {
-	    Tcl_EvalObjEx(interp, internalErrorScript, 
+	ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+	if (tsdPtr->internalErrorScript != NULL) {
+	    Tcl_EvalObjEx(interp, tsdPtr->internalErrorScript,
 			  TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
 	}
-	Tcl_MutexUnlock(&internalErrorMutex);
     }
 }
 
@@ -947,8 +944,8 @@ VfsFullyNormalizePath(Tcl_Interp *interp, Tcl_Obj *pathPtr) {
  *	internal representation for a vfs path.
  *	
  * Results:
- *	Returns TCL_OK on success, or '-1' on failure.  If Tcl is
- *	exiting, we always return a failure code.
+ *	Returns TCL_OK on success, or 'TCLVFS_POSIXERROR' on failure.
+ *	If Tcl is exiting, we always return a failure code.
  *
  * Side effects:
  *	On success, we allocate some memory for our internal
@@ -972,12 +969,12 @@ VfsPathInFilesystem(Tcl_Obj *pathPtr, ClientData *clientDataPtr) {
 	 * in the middle of the exit sequence.  We could perhaps be
 	 * more subtle than this!
 	 */
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     normedObj = Tcl_FSGetNormalizedPath(NULL, pathPtr);
     if (normedObj == NULL) {
-        return -1;
+        return TCLVFS_POSIXERROR;
     }
     normed = Tcl_GetStringFromObj(normedObj, &len);
     splitPosition = len;
@@ -1001,7 +998,7 @@ VfsPathInFilesystem(Tcl_Obj *pathPtr, ClientData *clientDataPtr) {
 	 * must return then.
 	 */
 	if (splitPosition == 0) {
-	    return -1;
+	    return TCLVFS_POSIXERROR;
 	}
 	
 	/* Is the path up to 'splitPosition' a valid moint point? */
@@ -1014,7 +1011,7 @@ VfsPathInFilesystem(Tcl_Obj *pathPtr, ClientData *clientDataPtr) {
 		 * We've reached the beginning of the string without
 		 * finding a mount, so we've failed.
 		 */
-		return -1;
+		return TCLVFS_POSIXERROR;
 	    }
 	}
 	
@@ -1105,7 +1102,7 @@ VfsStat(pathPtr, bufPtr)
     
     mountCmd = VfsBuildCommandForPath(&interp, "stat", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_SaveResult(interp, &savedResult);
@@ -1227,16 +1224,16 @@ VfsStat(pathPtr, bufPtr)
 	}
     }
     
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
 
     Tcl_RestoreResult(interp, &savedResult);
     Tcl_DecrRefCount(mountCmd);
     
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	Tcl_SetErrno(ENOENT);
-        return -1;
+        return TCLVFS_POSIXERROR;
     } else {
 	return returnVal;
     }
@@ -1254,7 +1251,7 @@ VfsAccess(pathPtr, mode)
     
     mountCmd = VfsBuildCommandForPath(&interp, "access", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(mode));
@@ -1262,7 +1259,7 @@ VfsAccess(pathPtr, mode)
     Tcl_SaveResult(interp, &savedResult);
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     Tcl_RestoreResult(interp, &savedResult);
@@ -1270,7 +1267,7 @@ VfsAccess(pathPtr, mode)
 
     if (returnVal != 0) {
 	Tcl_SetErrno(ENOENT);
-	return -1;
+	return TCLVFS_POSIXERROR;
     } else {
 	return returnVal;
     }
@@ -1353,7 +1350,7 @@ VfsOpenFileChannel(cmdInterp, pathPtr, mode, permissions)
     } else {
 	/* Leave an error message if the cmdInterp is non NULL */
 	if (cmdInterp != NULL) {
-	    if (returnVal == -1) {
+	    if (returnVal == TCLVFS_POSIXERROR) {
 		Tcl_ResetResult(cmdInterp);
 		Tcl_AppendResult(cmdInterp, "couldn't open \"", 
 				 Tcl_GetString(pathPtr), "\": ",
@@ -1368,7 +1365,7 @@ VfsOpenFileChannel(cmdInterp, pathPtr, mode, permissions)
 	    }
 	} else {
 	    /* Report any error, since otherwise it is lost */
-	    if (returnVal != -1) {
+	    if (returnVal != TCLVFS_POSIXERROR) {
 		VfsInternalError(interp);
 	    }
 	}
@@ -1393,8 +1390,23 @@ VfsOpenFileChannel(cmdInterp, pathPtr, mode, permissions)
 	 * Tcl_DetachChannel to do this for us.  We must use the
 	 * correct interpreter.
 	 */
+	if (Tcl_IsStandardChannel(chan)) {
+	    /*
+	     * If we have somehow ended up with a VFS channel being a std
+	     * channel, it is likely auto-inherited, which we need to reverse.
+	     * [Bug 1468291]
+	     */
+	    if (chan == Tcl_GetStdChannel(TCL_STDIN)) {
+		Tcl_SetStdChannel(NULL, TCL_STDIN);
+	    } else if (chan == Tcl_GetStdChannel(TCL_STDOUT)) {
+		Tcl_SetStdChannel(NULL, TCL_STDOUT);
+	    } else if (chan == Tcl_GetStdChannel(TCL_STDERR)) {
+		Tcl_SetStdChannel(NULL, TCL_STDERR);
+	    }
+	    Tcl_UnregisterChannel(NULL, chan);
+	}
 	Tcl_DetachChannel(interp, chan);
-	
+
 	if (closeCallback != NULL) {
 	    VfsChannelCleanupInfo *channelRet = NULL;
 	    channelRet = (VfsChannelCleanupInfo*) 
@@ -1439,7 +1451,9 @@ VfsCloseProc(ClientData clientData) {
      * callback will fail, so we register the channel (this allows
      * the Tcl code to use the channel's string-name).
      */
-    Tcl_RegisterChannel(interp, chan);
+    if (!Tcl_IsStandardChannel(chan)) {
+	Tcl_RegisterChannel(interp, chan);
+    }
 
     if (!(Tcl_GetChannelMode(chan) & TCL_READABLE)) {
 	/* 
@@ -1465,7 +1479,9 @@ VfsCloseProc(ClientData clientData) {
      * already being closed.  So, we do the same trick as above to
      * unregister it without cleanup.
      */
-    Tcl_DetachChannel(interp, chan);
+    if (!Tcl_IsStandardChannel(chan)) {
+	Tcl_DetachChannel(interp, chan);
+    }
 
     Tcl_RestoreResult(interp, &savedResult);
     ckfree((char*)channelRet);
@@ -1484,6 +1500,7 @@ VfsMatchInDirectory(
 	VfsMount *mountIter;
 	int len;
 	CONST char *prefix;
+	ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
 	prefix = Tcl_GetStringFromObj(Tcl_FSGetNormalizedPath(NULL, dirPtr), 
 				      &len);
@@ -1495,10 +1512,8 @@ VfsMatchInDirectory(
 	    len--;
 	}
 
-	Tcl_MutexLock(&vfsMountsMutex);
-
 	/* Build list of mounts */
-	mountIter = listOfMounts;
+	mountIter = tsdPtr->listOfMounts;
 	while (mountIter != NULL) {
 	    if (mountIter->mountLen > (len+1) 
 		&& !strncmp(mountIter->mountPoint, prefix, (size_t)len) 
@@ -1512,7 +1527,6 @@ VfsMatchInDirectory(
 	    }
 	    mountIter = mountIter->nextMount;
 	}
-	Tcl_MutexUnlock(&vfsMountsMutex);
 	return TCL_OK;
     } else {
 	Tcl_Obj *mountCmd = NULL;
@@ -1524,7 +1538,7 @@ VfsMatchInDirectory(
 	
 	mountCmd = VfsBuildCommandForPath(&interp, "matchindirectory", dirPtr);
 	if (mountCmd == NULL) {
-	    return -1;
+	    return TCLVFS_POSIXERROR;
 	}
 
 	if (types != NULL) {
@@ -1542,7 +1556,7 @@ VfsMatchInDirectory(
 	/* Now we execute this mount point's callback. */
 	returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 				  TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-	if (returnVal != -1) {
+	if (returnVal != TCLVFS_POSIXERROR) {
 	    vfsResultPtr = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
 	}
 	Tcl_RestoreResult(interp, &savedResult);
@@ -1554,7 +1568,11 @@ VfsMatchInDirectory(
 		Tcl_ListObjAppendList(cmdInterp, returnPtr, vfsResultPtr);
 		Tcl_DecrRefCount(vfsResultPtr);
 	    } else {
-		Tcl_SetObjResult(cmdInterp, vfsResultPtr);
+		if (cmdInterp != NULL) {
+		    Tcl_SetObjResult(cmdInterp, vfsResultPtr);
+		} else {
+		    Tcl_DecrRefCount(vfsResultPtr);
+		}
 	    }
 	}
 	return returnVal;
@@ -1572,14 +1590,14 @@ VfsDeleteFile(
     
     mountCmd = VfsBuildCommandForPath(&interp, "deletefile", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     /* Now we execute this mount point's callback. */
     Tcl_SaveResult(interp, &savedResult);
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     Tcl_RestoreResult(interp, &savedResult);
@@ -1598,14 +1616,14 @@ VfsCreateDirectory(
     
     mountCmd = VfsBuildCommandForPath(&interp, "createdirectory", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     /* Now we execute this mount point's callback. */
     Tcl_SaveResult(interp, &savedResult);
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     Tcl_RestoreResult(interp, &savedResult);
@@ -1630,7 +1648,7 @@ VfsRemoveDirectory(
     
     mountCmd = VfsBuildCommandForPath(&interp, "removedirectory", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(recursive));
@@ -1638,7 +1656,7 @@ VfsRemoveDirectory(
     Tcl_SaveResult(interp, &savedResult);
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     Tcl_RestoreResult(interp, &savedResult);
@@ -1655,7 +1673,7 @@ VfsRemoveDirectory(
     return returnVal;
 }
 
-static CONST char**
+static CONST char * CONST86 *
 VfsFileAttrStrings(pathPtr, objPtrRef)
     Tcl_Obj* pathPtr;
     Tcl_Obj** objPtrRef;
@@ -1675,7 +1693,7 @@ VfsFileAttrStrings(pathPtr, objPtrRef)
     /* Now we execute this mount point's callback. */
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     if (returnVal == TCL_OK) {
@@ -1702,7 +1720,7 @@ VfsFileAttrsGet(cmdInterp, index, pathPtr, objPtrRef)
     
     mountCmd = VfsBuildCommandForPath(&interp, "fileattributes", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(index));
@@ -1710,13 +1728,13 @@ VfsFileAttrsGet(cmdInterp, index, pathPtr, objPtrRef)
     /* Now we execute this mount point's callback. */
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != -1) {
+    if (returnVal != TCLVFS_POSIXERROR) {
 	*objPtrRef = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
     }
     Tcl_RestoreResult(interp, &savedResult);
     Tcl_DecrRefCount(mountCmd);
     
-    if (returnVal != -1) {
+    if (returnVal != TCLVFS_POSIXERROR) {
 	if (returnVal == TCL_OK) {
 	    /* 
 	     * Our caller expects a ref count of zero in
@@ -1758,7 +1776,7 @@ VfsFileAttrsSet(cmdInterp, index, pathPtr, objPtr)
     
     mountCmd = VfsBuildCommandForPath(&interp, "fileattributes", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(index));
@@ -1767,7 +1785,7 @@ VfsFileAttrsSet(cmdInterp, index, pathPtr, objPtr)
     /* Now we execute this mount point's callback. */
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != -1 && returnVal != TCL_OK) {
+    if (returnVal != TCLVFS_POSIXERROR && returnVal != TCL_OK) {
 	errorPtr = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
     }
 
@@ -1775,7 +1793,7 @@ VfsFileAttrsSet(cmdInterp, index, pathPtr, objPtr)
     Tcl_DecrRefCount(mountCmd);
     
     if (cmdInterp != NULL) {
-	if (returnVal == -1) {
+	if (returnVal == TCLVFS_POSIXERROR) {
 	    Tcl_ResetResult(cmdInterp);
 	    Tcl_AppendResult(cmdInterp, "couldn't set attributes for \"", 
 			     Tcl_GetString(pathPtr), "\": ",
@@ -1805,7 +1823,7 @@ VfsUtime(pathPtr, tval)
     
     mountCmd = VfsBuildCommandForPath(&interp, "utime", pathPtr);
     if (mountCmd == NULL) {
-	return -1;
+	return TCLVFS_POSIXERROR;
     }
 
     Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewLongObj(tval->actime));
@@ -1814,7 +1832,7 @@ VfsUtime(pathPtr, tval)
     Tcl_SaveResult(interp, &savedResult);
     returnVal = Tcl_EvalObjEx(interp, mountCmd, 
 			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != TCL_OK && returnVal != -1) {
+    if (returnVal != TCL_OK && returnVal != TCLVFS_POSIXERROR) {
 	VfsInternalError(interp);
     }
     Tcl_RestoreResult(interp, &savedResult);
@@ -1827,16 +1845,15 @@ static Tcl_Obj*
 VfsListVolumes(void)
 {
     Tcl_Obj *retVal;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
-    Tcl_MutexLock(&vfsVolumesMutex);
-    if (vfsVolumes != NULL) {
-	Tcl_IncrRefCount(vfsVolumes);
-	retVal = vfsVolumes;
+    if (tsdPtr->vfsVolumes != NULL) {
+	Tcl_IncrRefCount(tsdPtr->vfsVolumes);
+	retVal = tsdPtr->vfsVolumes;
     } else {
 	retVal = NULL;
     }
-    Tcl_MutexUnlock(&vfsVolumesMutex);
-    
+
     return retVal;
 }
 
@@ -1844,26 +1861,26 @@ static void
 Vfs_AddVolume(volume)
     Tcl_Obj *volume;
 {
-    Tcl_MutexLock(&vfsVolumesMutex);
-    
-    if (vfsVolumes == NULL) {
-        vfsVolumes = Tcl_NewObj();
-	Tcl_IncrRefCount(vfsVolumes);
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+
+    if (tsdPtr->vfsVolumes == NULL) {
+        tsdPtr->vfsVolumes = Tcl_NewObj();
+	Tcl_IncrRefCount(tsdPtr->vfsVolumes);
     } else {
-	if (Tcl_IsShared(vfsVolumes)) {
+#if 0
+	if (Tcl_IsShared(tsdPtr->vfsVolumes)) {
 	    /* 
 	     * Another thread is using this object, so we duplicate the
 	     * object and reduce the refCount on the shared one.
 	     */
-	    Tcl_Obj *oldVols = vfsVolumes;
-	    vfsVolumes = Tcl_DuplicateObj(oldVols);
-	    Tcl_IncrRefCount(vfsVolumes);
+	    Tcl_Obj *oldVols = tsdPtr->vfsVolumes;
+	    tsdPtr->vfsVolumes = Tcl_DuplicateObj(oldVols);
+	    Tcl_IncrRefCount(tsdPtr->vfsVolumes);
 	    Tcl_DecrRefCount(oldVols);
 	}
+#endif
     }
-    Tcl_ListObjAppendElement(NULL, vfsVolumes, volume);
-    
-    Tcl_MutexUnlock(&vfsVolumesMutex);
+    Tcl_ListObjAppendElement(NULL, tsdPtr->vfsVolumes, volume);
 }
 
 static int
@@ -1871,35 +1888,35 @@ Vfs_RemoveVolume(volume)
     Tcl_Obj *volume;
 {
     int i, len;
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
 
-    Tcl_MutexLock(&vfsVolumesMutex);
-
-    Tcl_ListObjLength(NULL, vfsVolumes, &len);
+    Tcl_ListObjLength(NULL, tsdPtr->vfsVolumes, &len);
     for (i = 0;i < len; i++) {
 	Tcl_Obj *vol;
-        Tcl_ListObjIndex(NULL, vfsVolumes, i, &vol);
+        Tcl_ListObjIndex(NULL, tsdPtr->vfsVolumes, i, &vol);
 	if (!strcmp(Tcl_GetString(vol),Tcl_GetString(volume))) {
 	    /* It's in the list, at index i */
 	    if (len == 1) {
 		/* An optimization here */
-		Tcl_DecrRefCount(vfsVolumes);
-		vfsVolumes = NULL;
+		Tcl_DecrRefCount(tsdPtr->vfsVolumes);
+		tsdPtr->vfsVolumes = NULL;
 	    } else {
-		/* Make ourselves the unique owner */
-		if (Tcl_IsShared(vfsVolumes)) {
-		    Tcl_Obj *oldVols = vfsVolumes;
-		    vfsVolumes = Tcl_DuplicateObj(oldVols);
-		    Tcl_IncrRefCount(vfsVolumes);
+		/*
+		 * Make ourselves the unique owner
+		 * XXX: May be unnecessary now that it is tsd
+		 */
+		if (Tcl_IsShared(tsdPtr->vfsVolumes)) {
+		    Tcl_Obj *oldVols = tsdPtr->vfsVolumes;
+		    tsdPtr->vfsVolumes = Tcl_DuplicateObj(oldVols);
+		    Tcl_IncrRefCount(tsdPtr->vfsVolumes);
 		    Tcl_DecrRefCount(oldVols);
 		}
 		/* Remove the element */
-		Tcl_ListObjReplace(NULL, vfsVolumes, i, 1, 0, NULL);
-		Tcl_MutexUnlock(&vfsVolumesMutex);
+		Tcl_ListObjReplace(NULL, tsdPtr->vfsVolumes, i, 1, 0, NULL);
 		return TCL_OK;
 	    }
 	}
     }
-    Tcl_MutexUnlock(&vfsVolumesMutex);
 
     return TCL_ERROR;
 }
@@ -2025,13 +2042,23 @@ static void
 VfsExitProc(ClientData clientData)
 {
     Tcl_FSUnregister(&vfsFilesystem);
-    /* 
+}
+
+static void
+VfsThreadExitProc(ClientData clientData)
+{
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+    /*
      * This is probably no longer needed, because each individual
      * interp's cleanup will trigger removal of all volumes which
      * belong to it.
      */
-    if (vfsVolumes != NULL) {
-        Tcl_DecrRefCount(vfsVolumes);
-	vfsVolumes = NULL;
+    if (tsdPtr->vfsVolumes != NULL) {
+	Tcl_DecrRefCount(tsdPtr->vfsVolumes);
+	tsdPtr->vfsVolumes = NULL;
+    }
+    if (tsdPtr->internalErrorScript != NULL) {
+	Tcl_DecrRefCount(tsdPtr->internalErrorScript);
+	tsdPtr->internalErrorScript = NULL;
     }
 }
